"""Unit tests for :mod:`~tmlt.core.measurements.interactive_measurements`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

# pylint: disable=no-self-use

from typing import Any, List, Optional, Tuple, Type, Union
from unittest import TestCase
from unittest.mock import ANY, MagicMock, Mock, patch

import numpy as np
import sympy as sp
from parameterized import parameterized, parameterized_class

from tmlt.core.domains.base import Domain
from tmlt.core.domains.collections import DictDomain, ListDomain
from tmlt.core.domains.numpy_domains import (
    NumpyFloatDomain,
    NumpyIntegerDomain,
    NumpyStringDomain,
)
from tmlt.core.domains.spark_domains import (
    SparkDataFrameDomain,
    SparkIntegerColumnDescriptor,
    SparkStringColumnDescriptor,
)
from tmlt.core.exceptions import OutOfDomainError, UnsupportedDomainError
from tmlt.core.measurements.base import Measurement
from tmlt.core.measurements.interactive_measurements import (
    DecoratedQueryable,
    DecorateQueryable,
    GetAnswerQueryable,
    IndexQuery,
    MakeInteractive,
    MeasurementQuery,
    ParallelComposition,
    ParallelQueryable,
    PrivacyAccountant,
    PrivacyAccountantState,
    Queryable,
    RetirableQueryable,
    RetireQuery,
    SequentialComposition,
    SequentialQueryable,
    TransformationQuery,
    create_adaptive_composition,
)
from tmlt.core.measures import ApproxDP, Measure, PureDP, RhoZCDP
from tmlt.core.metrics import (
    AbsoluteDifference,
    DictMetric,
    HammingDistance,
    Metric,
    RootSumOfSquared,
    SumOf,
    SymmetricDifference,
)
from tmlt.core.transformations.base import Transformation
from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput
from tmlt.core.utils.testing import (
    PySparkTest,
    assert_property_immutability,
    create_mock_measurement,
    create_mock_queryable,
    create_mock_transformation,
    get_all_props,
)


@parameterized_class(
    [
        {"output_measure": PureDP(), "privacy_budget": 6},
        {"output_measure": ApproxDP(), "privacy_budget": (6, sp.Rational("0.1"))},
        {"output_measure": RhoZCDP(), "privacy_budget": 6},
    ]
)
class TestSequentialComposition(PySparkTest):
    """Tests for :class:`~.SequentialComposition`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    privacy_budget: Union[int, Tuple[int, sp.Rational]]
    """The privacy budget to use during the tests."""

    def setUp(self):
        """Set up class."""
        self.data = np.int64(10)
        self.measurement = SequentialComposition(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            output_measure=self.output_measure,
            d_in=1,
            privacy_budget=self.privacy_budget,
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        d_in = {"A": 2}
        measurement = SequentialComposition(
            input_domain=DictDomain({"A": NumpyIntegerDomain()}),
            input_metric=DictMetric({"A": AbsoluteDifference()}),
            output_measure=self.output_measure,
            d_in=d_in,
            privacy_budget=self.privacy_budget,
        )
        d_in["A"] = 1
        d_in["B"] = 2
        self.assertDictEqual(measurement.d_in, {"A": 2})

    @parameterized.expand(get_all_props(SequentialComposition))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        measurement = SequentialComposition(
            input_domain=DictDomain({"A": NumpyIntegerDomain()}),
            input_metric=DictMetric({"A": AbsoluteDifference()}),
            output_measure=self.output_measure,
            d_in={"A": 1},
            privacy_budget=self.privacy_budget,
        )
        assert_property_immutability(measurement, prop_name)

    def test_properties(self):
        """SequentialComposition's properties have the expected values."""
        self.assertEqual(self.measurement.input_domain, NumpyIntegerDomain())
        self.assertEqual(self.measurement.input_metric, AbsoluteDifference())
        self.assertEqual(self.measurement.output_measure, self.output_measure)
        self.assertEqual(self.measurement.is_interactive, True)
        self.assertEqual(self.measurement.d_in, 1)
        self.assertEqual(self.measurement.privacy_budget, self.privacy_budget)

    def test_privacy_function(self):
        """SequentialComposition's privacy function is correct."""
        self.assertEqual(self.measurement.privacy_function(1), self.privacy_budget)

    def test_privacy_function_invalid_d_in(self):
        """SequentialComposition's privacy function raises error if d_in is invalid."""
        with self.assertRaisesRegex(ValueError, "d_in must be <= 1, not 2"):
            self.measurement.privacy_function(2)

    def test_correctness(self):
        """SequentialComposition returns the expected Queryable object."""
        # pylint: disable=protected-access
        actual = self.measurement(self.data)
        self.assertIsInstance(actual, SequentialQueryable)
        self.assertEqual(actual._input_domain, self.measurement.input_domain)
        self.assertEqual(actual._input_metric, self.measurement.input_metric)
        self.assertEqual(actual._output_measure, self.measurement.output_measure)
        self.assertEqual(
            actual._remaining_budget.value, self.measurement.privacy_budget
        )
        self.assertEqual(actual._data, self.data)


@parameterized_class(
    [
        {
            "output_measure": PureDP(),
            "input_metric": SumOf(AbsoluteDifference()),
            "composed_measurements": [
                create_mock_measurement(
                    output_measure=PureDP(),
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=ExactNumber(6),
                )
                for _ in range(3)
            ],
            "privacy_budget": 6,
        },
        {
            "output_measure": ApproxDP(),
            "input_metric": SumOf(AbsoluteDifference()),
            "composed_measurements": [
                create_mock_measurement(
                    output_measure=ApproxDP(),
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=(
                        ExactNumber(6),
                        ExactNumber(sp.Rational("0.1")),
                    ),
                ),
                create_mock_measurement(
                    output_measure=ApproxDP(),
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=(
                        ExactNumber(2),
                        ExactNumber(sp.Rational("0.3")),
                    ),
                ),
                create_mock_measurement(
                    output_measure=ApproxDP(),
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=(
                        ExactNumber(1),
                        ExactNumber(sp.Rational("0.2")),
                    ),
                ),
            ],
            "privacy_budget": (6, sp.Rational("0.3")),
        },
        {
            "output_measure": RhoZCDP(),
            "input_metric": RootSumOfSquared(AbsoluteDifference()),
            "composed_measurements": [
                create_mock_measurement(
                    output_measure=RhoZCDP(),
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=ExactNumber(6),
                )
                for _ in range(3)
            ],
            "privacy_budget": 6,
        },
    ]
)
class TestParallelComposition(PySparkTest):
    """Tests for :class:`~.ParallelComposition`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    composed_measurements: List[Measurement]
    """The measurements to compose."""

    privacy_budget: Union[int, Tuple[int, sp.Rational]]
    """The privacy budget to use during the tests."""

    input_metric: Union[SumOf, RootSumOfSquared]
    """The input metric to use during the tests."""

    def setUp(self):
        """Test setup."""
        self.input_domain = ListDomain(NumpyIntegerDomain(), length=3)
        self.measurement = ParallelComposition(
            input_domain=self.input_domain,
            input_metric=self.input_metric,
            output_measure=self.output_measure,
            measurements=self.composed_measurements,
        )

        self.data = [np.int64(10), np.int64(30), np.int64(20)]

    @parameterized.expand(get_all_props(ParallelComposition))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.measurement, prop_name)

    @parameterized.expand(
        [
            (
                ListDomain(NumpyIntegerDomain(), length=1),
                SumOf(AbsoluteDifference()),
                PureDP(),
                [create_mock_measurement(is_interactive=False)],
                "All measurements must be interactive",
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                PureDP(),
                [create_mock_measurement(is_interactive=True)],
                r"Length of input domain \(2\) does not match the number of"
                r" measurements \(1\)",
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=2),
                RootSumOfSquared(AbsoluteDifference()),
                ApproxDP(),
                [create_mock_measurement(is_interactive=True)],
                (
                    f"Input metric {RootSumOfSquared} is incompatible with output"
                    f" measure {ApproxDP}"
                ),
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=1),
                SumOf(AbsoluteDifference()),
                RhoZCDP(),
                [create_mock_measurement(is_interactive=True)],
                f"Input metric {SumOf} is incompatible with output measure {RhoZCDP}",
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=1),
                RootSumOfSquared(AbsoluteDifference()),
                RhoZCDP(),
                [
                    create_mock_measurement(
                        is_interactive=True,
                        output_measure=RhoZCDP(),
                        input_metric=SymmetricDifference(),
                    )
                ],
                (
                    "Input metric for each supplied measurement must match inner metric"
                    " of input metric for ParallelComposition"
                ),
            ),
            (
                ListDomain(NumpyIntegerDomain(), length=1),
                RootSumOfSquared(AbsoluteDifference()),
                RhoZCDP(),
                [create_mock_measurement(is_interactive=True, output_measure=PureDP())],
                (
                    "Output measure for each supplied measurement must match output"
                    " measure for ParallelComposition"
                ),
            ),
        ]
    )
    def test_invalid_constructor_arguments(
        self,
        input_domain: ListDomain,
        input_metric: Union[SumOf, RootSumOfSquared],
        output_measure: Union[PureDP, RhoZCDP],
        measurements: List[Measurement],
        expected_error_message: str,
    ):
        """ParallelComposition constructor raises appropriate error on invalid args."""
        with self.assertRaisesRegex(
            (ValueError, OutOfDomainError), expected_error_message
        ):
            ParallelComposition(
                input_domain=input_domain,
                input_metric=input_metric,
                output_measure=output_measure,
                measurements=measurements,
            )

    def test_properties(self):
        """ParallelComposition's properties have the expected values."""
        self.assertEqual(self.measurement.input_domain, self.input_domain)
        self.assertEqual(self.measurement.input_metric, self.input_metric)
        self.assertEqual(self.measurement.output_measure, self.output_measure)
        self.assertEqual(self.measurement.is_interactive, True)

    def test_privacy_function(self):
        """ParallelComposition's privacy function is correct."""
        self.assertEqual(self.measurement.privacy_function(1), self.privacy_budget)

    def test_correctness(self):
        """ParallelComposition returns the expected Queryable object."""
        # pylint: disable=protected-access
        actual = self.measurement(self.data)
        self.assertIsInstance(actual, ParallelQueryable)
        self.assertEqual(actual._next_index, 0)
        self.assertEqual(actual._data, self.data)
        self.assertEqual(actual._measurements, self.composed_measurements)


class TestMakeInteractive(PySparkTest):
    """Tests for :class:`~.MakeInteractive`."""

    def test_interactive_measurement_raises_error(self):
        """MakeInteractive can not be constructed with an interactive measurement."""
        with self.assertRaisesRegex(ValueError, "Measurement must be non-interactive"):
            MakeInteractive(measurement=create_mock_measurement(is_interactive=True))

    @parameterized.expand(get_all_props(MakeInteractive))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        measurement = MakeInteractive(measurement=create_mock_measurement())
        assert_property_immutability(measurement, prop_name)

    def test_privacy_function(self):
        """MakeInteractive has correct privacy_function."""
        non_interactive_measurement = create_mock_measurement(
            privacy_function_implemented=True, privacy_function_return_value=20
        )
        actual = MakeInteractive(non_interactive_measurement).privacy_function(d_in=12)
        self.assertEqual(actual, 20)
        non_interactive_measurement.privacy_function.assert_called_once_with(d_in=12)

    def test_correctness(self):
        """MakeInteractive returns expected GetAnswerQueryable."""
        non_interactive_measurement = create_mock_measurement(return_value=2)
        measurement = MakeInteractive(non_interactive_measurement)
        queryable = measurement(None)
        self.assertIsInstance(queryable, GetAnswerQueryable)
        answer = queryable(None)
        self.assertEqual(answer, 2)

    def test_properties(self):
        """MakeInteractive's properties have appropriate values."""
        non_interactive_measurement = create_mock_measurement()
        measurement = MakeInteractive(non_interactive_measurement)
        self.assertEqual(
            measurement.input_domain, non_interactive_measurement.input_domain
        )
        self.assertEqual(
            measurement.input_metric, non_interactive_measurement.input_metric
        )
        self.assertEqual(
            measurement.output_measure, non_interactive_measurement.output_measure
        )
        self.assertTrue(measurement.is_interactive)
        self.assertEqual(measurement.measurement, non_interactive_measurement)


class TestRetirableQueryable(PySparkTest):
    """Tests for :class:`~.RetirableQueryable`."""

    def test_retirable_queryable_returned(self):
        """Returned Queryable is wrapped in RetirableQueryable."""
        returned_queryable = create_mock_queryable()
        queryable = RetirableQueryable(
            create_mock_queryable(return_value=returned_queryable)
        )
        actual = queryable(None)
        self.assertIsInstance(actual, RetirableQueryable)

        self.assertEqual(
            actual._inner_queryable,  # pylint: disable=protected-access
            returned_queryable,
        )

    def test_retire_works_recursively(self):
        """RetireQuery is propagated to all descendants correctly."""
        queryable = RetirableQueryable(
            create_mock_queryable(
                return_value=create_mock_queryable(return_value=create_mock_queryable())
            )
        )
        inner_most_queryable = queryable(None)(None)
        queryable(RetireQuery())
        self.assertTrue(
            inner_most_queryable._is_retired  # pylint: disable=protected-access
        )

    def test_retire_works_when_descendant_is_retired(self):
        """RetirableQueryable can be retired even when a descendant is retired."""
        queryable = RetirableQueryable(
            create_mock_queryable(
                return_value=create_mock_queryable(return_value=create_mock_queryable())
            )
        )
        child_queryable = queryable(None)
        _ = child_queryable(None)
        child_queryable(RetireQuery())  # Retired descendants
        queryable(RetireQuery())

    def test_retired_retirable_raises_error(self):
        """Retired RetirableQueryables raise error when a query is submitted."""
        queryable = RetirableQueryable(create_mock_queryable())
        queryable(RetireQuery())  # Retired descendants
        with self.assertRaisesRegex(ValueError, "Queryable already retired"):
            queryable(None)


@parameterized_class(
    [
        {
            "output_measure": PureDP(),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
        {
            "output_measure": ApproxDP(),
            "budget_quarters": [
                (0, 0),
                (sp.Rational("2.5"), sp.Rational("0.1")),
                (5, sp.Rational("0.2")),
                (sp.Rational("7.5"), sp.Rational("0.3")),
                (10, sp.Rational("0.4")),
            ],
        },
        {
            "output_measure": RhoZCDP(),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
    ]
)
class TestSequentialQueryable(PySparkTest):
    """Tests for :class:`~.SequentialQueryable`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    budget_quarters: List
    """Zero, one quarter, one half, three quarters, and all of the privacy budget."""

    def setUp(self):
        """Test setup."""
        self.privacy_budget = self.budget_quarters[4]
        self.construct_queryable = lambda: SequentialQueryable(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            d_in=1,
            output_measure=self.output_measure,
            privacy_budget=self.privacy_budget,
            data=np.int64(10),
        )

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        d_in = {"A": 2}
        queryable = SequentialQueryable(
            input_domain=DictDomain({"A": NumpyIntegerDomain()}),
            input_metric=DictMetric({"A": AbsoluteDifference()}),
            d_in=d_in,
            output_measure=PureDP(),
            privacy_budget=10,
            data={"A": np.int64(10)},
        )
        d_in["A"] = 1
        d_in["B"] = 2
        self.assertDictEqual(
            queryable._d_in, {"A": 2}  # pylint: disable=protected-access
        )

    def test_queryable_budget_is_decreased_correctly(self):
        """SequentialQueryable's internal budget is correctly decreased on query."""
        queryable = self.construct_queryable()
        queryable(
            MeasurementQuery(
                measurement=create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=self.budget_quarters[1],
                    return_value=create_mock_queryable(),
                )
            )
        )
        self.assertEqual(
            queryable._remaining_budget.value,  # pylint: disable=protected-access
            self.budget_quarters[3],
        )

    def test_insufficient_budget_remaining(self):
        """Error is raised when query privacy loss exceeds available budget."""
        queryable = self.construct_queryable()
        queryable(
            MeasurementQuery(
                measurement=create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    privacy_function_implemented=True,
                    privacy_function_return_value=self.budget_quarters[2],
                    return_value=create_mock_queryable(),
                )
            )
        )
        with self.assertRaisesRegex(
            ValueError, "Cannot answer query without exceeding available privacy budget"
        ):
            queryable(
                MeasurementQuery(
                    measurement=create_mock_measurement(
                        output_measure=self.output_measure,
                        is_interactive=True,
                        privacy_function_implemented=True,
                        privacy_function_return_value=self.budget_quarters[3],
                    )
                )
            )

    def test_non_interactive_measurement_disallowed(self):
        """SequentialQueryable raises error when measurement is non-interactive."""
        queryable = self.construct_queryable()
        with self.assertRaisesRegex(
            ValueError,
            "SequentialQueryable does not answer non-interactive measurement",
        ):
            queryable(
                MeasurementQuery(
                    create_mock_measurement(output_measure=self.output_measure)
                )
            )

    @parameterized.expand(
        [
            (
                NumpyFloatDomain(),
                AbsoluteDifference(),
                PureDP(),
                (
                    "Input domain of measurement query does not match the input domain"
                    " of SequentialQueryable."
                ),
            ),
            (
                NumpyIntegerDomain(),
                SymmetricDifference(),
                PureDP(),
                (
                    "Input metric of measurement query does not match the input metric"
                    " of SequentialQueryable."
                ),
            ),
            (
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                RhoZCDP(),
                (
                    "Output measure of measurement query does not match the output"
                    " measure of SequentialQueryable."
                ),
            ),
        ]
    )
    def test_incompatible_measurement_query(
        self,
        input_domain: Domain,
        input_metric: Metric,
        measure: Measure,
        expected_error_message: str,
    ):
        """SequentialQueryable raises error if query is incompatible."""
        queryable = SequentialQueryable(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            d_in=1,
            output_measure=PureDP(),
            privacy_budget=10,
            data=np.int64(10),
        )
        with self.assertRaisesRegex(ValueError, expected_error_message):
            queryable(
                MeasurementQuery(
                    create_mock_measurement(
                        is_interactive=True,
                        input_domain=input_domain,
                        input_metric=input_metric,
                        output_measure=measure,
                    )
                )
            )

    @parameterized.expand([(True,), (False,)])
    def test_measurement_query_with_explicit_d_out(
        self, privacy_function_implemented: bool
    ):
        """SequentialQueryable can run transformations without stability function."""
        queryable = self.construct_queryable()
        queryable(
            MeasurementQuery(
                create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    privacy_function_implemented=privacy_function_implemented,
                    privacy_relation_return_value=True,
                    privacy_function_return_value=self.budget_quarters[1],
                    return_value=create_mock_queryable(),
                ),
                d_out=self.budget_quarters[2],
            )
        )
        self.assertEqual(
            queryable._remaining_budget.value,  # pylint: disable=protected-access
            self.budget_quarters[2],
        )

    def test_measurement_query_stability_relation_returns_false(self):
        """SequentialQueryable raises error if privacy relation is not True."""
        queryable = self.construct_queryable()
        with self.assertRaisesRegex(
            ValueError,
            "Measurement's privacy relation cannot be satisfied with given d_out",
        ):
            queryable(
                MeasurementQuery(
                    create_mock_measurement(
                        output_measure=self.output_measure,
                        is_interactive=True,
                        privacy_function_implemented=False,
                        privacy_relation_return_value=False,
                    ),
                    d_out=self.privacy_budget,
                )
            )

    @parameterized.expand([(True,), (False,)])
    def test_transformation_query_with_explicit_d_out(
        self, stability_function_implemented: bool
    ):
        """SequentialQueryable can run transformations without stability function."""
        queryable = self.construct_queryable()
        queryable(
            TransformationQuery(
                create_mock_transformation(
                    stability_function_implemented=stability_function_implemented,
                    stability_function_return_value=2,
                    stability_relation_return_value=True,
                ),
                d_out=3,
            )
        )
        self.assertEqual(queryable._d_in, 3)  # pylint: disable=protected-access

    def test_transformation_query_stability_relation_returns_false(self):
        """SequentialQueryable raises error if stability relation is not True."""
        queryable = self.construct_queryable()
        with self.assertRaisesRegex(
            ValueError,
            "Transformation's stability relation cannot be satisfied with given"
            r" d_out \(3\)",
        ):
            queryable(
                TransformationQuery(
                    create_mock_transformation(
                        stability_function_implemented=False,
                        stability_relation_return_value=False,
                    ),
                    d_out=3,
                )
            )

    def test_transformation_query(self):
        """SequentialQueryable processes TransformationQuery correctly."""
        # pylint: disable=protected-access
        queryable = self.construct_queryable()
        transformation = create_mock_transformation(
            return_value=np.float64(100.0),
            stability_function_implemented=True,
            output_domain=NumpyFloatDomain(),
            output_metric=SymmetricDifference(),
            stability_function_return_value=20,
        )
        queryable(TransformationQuery(transformation=transformation))
        self.assertEqual(queryable._d_in, 20)
        self.assertEqual(queryable._input_domain, NumpyFloatDomain())
        self.assertEqual(queryable._input_metric, SymmetricDifference())
        self.assertEqual(queryable._data, np.float64(100.0))

    @parameterized.expand(
        [
            (
                NumpyIntegerDomain(),
                SymmetricDifference(),
                (
                    "Input metric of transformation query does not match the input"
                    " metric of SequentialQueryable."
                ),
            ),
            (
                NumpyFloatDomain(),
                AbsoluteDifference(),
                (
                    "Input domain of transformation query does not match the input"
                    " domain of SequentialQueryable."
                ),
            ),
        ]
    )
    def test_invalid_transformation_query(
        self, input_domain: Domain, input_metric: Metric, error_message: str
    ):
        """SequentialQueryable raises error when TransformationQuery is incompatible."""
        queryable = self.construct_queryable()
        transformation = create_mock_transformation(
            stability_function_implemented=True,
            stability_function_return_value=2,
            input_domain=input_domain,
            input_metric=input_metric,
        )
        with self.assertRaisesRegex(ValueError, error_message):
            queryable(TransformationQuery(transformation=transformation))

    def test_interactive_measurements_return_retirables(self):
        """SequentialQueryable returns a RetirableQueryable when answering queries."""
        queryable = self.construct_queryable()
        answer = queryable(
            MeasurementQuery(
                create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    return_value=create_mock_queryable(),
                    privacy_function_implemented=True,
                    privacy_function_return_value=self.privacy_budget,
                )
            )
        )
        self.assertIsInstance(answer, RetirableQueryable)

    def test_interleaving_is_disallowed(self):
        """Tests that interleaving is not allowed."""
        queryable = SequentialQueryable(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            d_in=1,
            output_measure=self.output_measure,
            privacy_budget=self.privacy_budget,
            data=np.int64(10),
        )
        child1_queryable = queryable(
            MeasurementQuery(
                create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    privacy_function_implemented=True,
                    return_value=create_mock_queryable(
                        return_value=create_mock_queryable()
                    ),
                    privacy_function_return_value=self.budget_quarters[1],
                )
            )
        )
        grandchild_queryable = child1_queryable(None)
        _ = queryable(
            MeasurementQuery(
                create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    return_value=create_mock_queryable(),
                    privacy_function_implemented=True,
                    privacy_function_return_value=self.budget_quarters[1],
                )
            )
        )
        with self.assertRaisesRegex(ValueError, "Queryable already retired"):
            grandchild_queryable(None)


@parameterized_class(
    [
        {"output_measure": PureDP()},
        {"output_measure": ApproxDP()},
        {"output_measure": RhoZCDP()},
    ]
)
class TestParallelQueryable(PySparkTest):
    """Tests for :class:`~.ParallelQueryable`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    def setUp(self):
        """Test setup."""
        self.construct_queryable = lambda: ParallelQueryable(
            data=[np.int64(10), np.int64(1)],
            measurements=[
                create_mock_measurement(
                    output_measure=self.output_measure,
                    is_interactive=True,
                    return_value=create_mock_queryable(
                        return_value=create_mock_queryable()
                    ),
                )
                for _ in range(2)
            ],
        )

    def test_invalid_constructor_arguments(self):
        """ParallelQueryable raises an error when arguments are invalid."""
        with self.assertRaisesRegex(
            ValueError,
            "Length of input data does not match the number of measurements provided",
        ):
            ParallelQueryable(
                data=[np.int64(10), np.int64(1)],
                measurements=[
                    create_mock_measurement(
                        is_interactive=True,
                        return_value=create_mock_queryable(
                            return_value=create_mock_queryable()
                        ),
                    )
                ],
            )

    def test_returned_queryables_are_retirable(self):
        """Index queries are answered with RetirableQueryables."""
        queryable = self.construct_queryable()
        self.assertIsInstance(queryable(IndexQuery(0)), RetirableQueryable)

    def test_sequential_access_enforced(self):
        """ParallelQueryable raises error if access is not in expected order."""
        queryable = self.construct_queryable()
        with self.assertRaisesRegex(ValueError, "Bad Index"):
            queryable(IndexQuery(1))

    def test_interleaving_is_disallowed(self):
        """ParallelQueryable disallows interleaving."""
        queryable = self.construct_queryable()
        child1_queryable = queryable(IndexQuery(0))
        grandchild_queryable = child1_queryable(None)
        _ = queryable(IndexQuery(1))
        with self.assertRaisesRegex(ValueError, "Queryable already retired"):
            grandchild_queryable(None)


class TestGetAnswerQueryable(PySparkTest):
    """Tests for :class:`~.GetAnswerQueryable`."""

    def test_disallows_interactive_measurement(self):
        """GetAnswerQueryable cannot be constructed with an interactive measurement."""
        with self.assertRaisesRegex(ValueError, "Measurement must be non-interactive"):
            GetAnswerQueryable(
                create_mock_measurement(is_interactive=True), data=np.int64(10)
            )

    def test_correctness(self):
        """GetAnswerQueryable returns measurement answer as expected."""
        non_interactive_measurement = create_mock_measurement()
        self.assertEqual(
            GetAnswerQueryable(non_interactive_measurement, np.int64(10))(None),
            non_interactive_measurement(np.int64(10)),
        )


@parameterized_class(
    [
        {
            "output_measure": PureDP(),
            "splitting_output_metric": SumOf(AbsoluteDifference()),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
        {
            "output_measure": ApproxDP(),
            "splitting_output_metric": SumOf(AbsoluteDifference()),
            "budget_quarters": [
                (0, 0),
                (sp.Rational("2.5"), sp.Rational("0.1")),
                (5, sp.Rational("0.2")),
                (sp.Rational("7.5"), sp.Rational("0.3")),
                (10, sp.Rational("0.4")),
            ],
        },
        {
            "output_measure": RhoZCDP(),
            "splitting_output_metric": RootSumOfSquared(AbsoluteDifference()),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
    ]
)
class TestPrivacyAccountant(PySparkTest):
    """Tests for :class:`~.PrivacyAccountant`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    splitting_output_metric: Union[SumOf, RootSumOfSquared]
    """The output metric for splitting transformations."""

    budget_quarters: List
    """Zero, one quarter, one half, three quarters, and all of the privacy budget."""

    def setUp(self):
        """Test Setup."""
        self.privacy_budget = self.budget_quarters[4]
        self.data = self.spark.createDataFrame(
            [(1, "X"), (2, "Y"), (3, "Z")], schema=["A", "B"]
        )
        self.measurement = SequentialComposition(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_measure=self.output_measure,
            d_in=1,
            privacy_budget=self.privacy_budget,
        )
        self.get_queryable = lambda: self.measurement(self.data)

    def test_constructor_mutable_arguments(self):
        """Tests that mutable constructor arguments are copied."""
        d_in = {"A": 2}
        accountant = PrivacyAccountant(
            queryable=SequentialQueryable(
                input_domain=DictDomain({"A": NumpyIntegerDomain()}),
                input_metric=DictMetric({"A": AbsoluteDifference()}),
                d_in=d_in,
                output_measure=self.output_measure,
                privacy_budget=self.privacy_budget,
                data={"A": np.int64(10)},
            ),
            input_metric=DictMetric({"A": AbsoluteDifference()}),
            input_domain=DictDomain({"A": NumpyIntegerDomain()}),
            output_measure=self.output_measure,
            d_in=d_in,
            privacy_budget=self.privacy_budget,
        )
        d_in["A"] = 1
        d_in["B"] = 2
        self.assertDictEqual(accountant.d_in, {"A": 2})

    def test_children_cannot_be_mutated(self):
        """Tests that the list of children of a PrivacyAccountant cannot be mutated."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        children = accountant.split(
            splitting_transformation=create_mock_transformation(
                input_domain=self.measurement.input_domain,
                input_metric=self.measurement.input_metric,
                output_domain=ListDomain(NumpyIntegerDomain(), length=4),
                output_metric=self.splitting_output_metric,
                return_value=[np.int64(0) for _ in range(4)],
                stability_function_implemented=True,
            ),
            privacy_budget=self.budget_quarters[1],
        )
        self.assertEqual(children, accountant.children)
        children.append(accountant)
        self.assertNotEqual(children, accountant.children)

        children = accountant.children
        self.assertEqual(children, accountant.children)
        children.append(accountant)
        self.assertNotEqual(children, accountant.children)

    @parameterized.expand(
        [
            (
                None,
                None,
                (
                    "PrivacyAccountant cannot be initialized with no parent and no"
                    " queryable"
                ),
            ),
            (
                Mock(PrivacyAccountant),
                Mock(SequentialQueryable),
                (
                    "PrivacyAccountant can be initialized with only parent or only"
                    " queryable but not both"
                ),
            ),
        ]
    )
    def test_init_invalid_arguments(
        self,
        parent: Optional[PrivacyAccountant],
        queryable: Optional[Queryable],
        error_message: str,
    ):
        """PrivacyAccountant.__init__ raises error when arguments are invalid."""
        with self.assertRaisesRegex(ValueError, error_message):
            PrivacyAccountant(
                queryable=queryable,
                parent=parent,
                input_metric=AbsoluteDifference(),
                input_domain=NumpyIntegerDomain(),
                output_measure=self.output_measure,
                d_in=1,
                privacy_budget=self.privacy_budget,
            )

    @patch.object(PrivacyAccountant, "__init__", autospec=True, return_value=None)
    def test_launch(self, mock_accountant_init):
        """PrivacyAccountant.launch works as expected."""

        mock_queryable = Mock(spec=SequentialQueryable)
        mock_sequential_composition = Mock(
            spec=SequentialComposition, return_value=mock_queryable
        )
        mock_sequential_composition.d_in = 1
        mock_sequential_composition.privacy_budget = self.privacy_budget
        mock_sequential_composition.input_domain = NumpyIntegerDomain()
        mock_sequential_composition.input_metric = AbsoluteDifference()
        mock_sequential_composition.output_measure = self.output_measure

        PrivacyAccountant.launch(
            measurement=mock_sequential_composition, data=self.data
        )
        mock_accountant_init.assert_called_with(
            self=ANY,
            queryable=mock_queryable,
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            output_measure=self.output_measure,
            d_in=1,
            privacy_budget=self.privacy_budget,
        )

    def test_privacy_accountant_properties(self):
        """PrivacyAccountant's properties have expected values."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        self.assertEqual(accountant.state, PrivacyAccountantState.ACTIVE)
        self.assertEqual(accountant.input_domain, self.measurement.input_domain)
        self.assertEqual(accountant.input_metric, self.measurement.input_metric)
        self.assertEqual(accountant.output_measure, self.measurement.output_measure)
        self.assertEqual(accountant.privacy_budget, self.measurement.privacy_budget)
        self.assertEqual(accountant.d_in, self.measurement.d_in)
        self.assertEqual(accountant.children, [])

    @parameterized.expand(
        [
            (
                (
                    "Transformation's input domain does not match PrivacyAccountant's"
                    " input domain"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {"A": SparkIntegerColumnDescriptor()}
                    ),
                    input_metric=SymmetricDifference(),
                    stability_function_implemented=True,
                ),
            ),
            (
                (
                    "Transformation's input metric does not match PrivacyAccountant's"
                    " input metric"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {
                            "A": SparkIntegerColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                        }
                    ),
                    input_metric=HammingDistance(),
                    stability_function_implemented=True,
                ),
            ),
        ]
    )
    def test_transform_in_place_invalid_arguments(
        self,
        error_message: str,
        transformation: Transformation,
        d_out: Optional[Any] = None,
    ):
        """PrivacyAccountant raises error transformation can not be applied."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        with self.assertRaisesRegex(ValueError, error_message):
            accountant.transform_in_place(transformation=transformation, d_out=d_out)

    def test_transform_in_place(self):
        """PrivacyAccountant.transform_in_place works correctly."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=NumpyIntegerDomain(),
            output_metric=AbsoluteDifference(),
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=np.int64(2),
        )
        accountant.transform_in_place(transformation=transformation)
        self.assertEqual(accountant.input_domain, NumpyIntegerDomain())
        self.assertEqual(accountant.input_metric, AbsoluteDifference())
        self.assertEqual(accountant.d_in, 10)
        # pylint: disable=protected-access
        self.assertIsNotNone(accountant._queryable)
        self.assertEqual(accountant._queryable._data, np.int64(2))  # type: ignore
        # pylint: enable=protected-access

    def test_transform_with_explicit_d_out(self):
        """PrivacyAccountant.transform_in_place works with a d_out provided."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=NumpyIntegerDomain(),
            output_metric=AbsoluteDifference(),
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=np.int64(2),
        )
        accountant.transform_in_place(transformation=transformation, d_out=10)
        self.assertEqual(accountant.input_domain, NumpyIntegerDomain())
        self.assertEqual(accountant.input_metric, AbsoluteDifference())
        self.assertEqual(accountant.d_in, 10)
        # pylint: disable=protected-access
        self.assertIsNotNone(accountant._queryable)
        self.assertEqual(accountant._queryable._data, np.int64(2))  # type: ignore

    @parameterized.expand(
        [
            (
                (
                    "Measurement's output measure does not match PrivacyAccountant's"
                    " output measure"
                ),
                create_mock_measurement(
                    input_domain=SparkDataFrameDomain(
                        {
                            "A": SparkIntegerColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                        }
                    ),
                    input_metric=SymmetricDifference(),
                    output_measure=RhoZCDP(),
                    privacy_function_implemented=True,
                    privacy_function_return_value=5,
                ),
            )
        ]
    )
    def test_measure_invalid_arguments(
        self,
        error_message: str,
        measurement: Measurement,
        error_type: Type = ValueError,
    ):
        """PrivacyAccountant raises error when a measurement cannot be answered."""
        accountant = PrivacyAccountant.launch(
            measurement=SequentialComposition(
                input_domain=SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(),
                        "B": SparkStringColumnDescriptor(),
                    }
                ),
                input_metric=SymmetricDifference(),
                output_measure=PureDP(),
                d_in=1,
                privacy_budget=6,
            ),
            data=self.data,
        )
        with self.assertRaisesRegex(error_type, error_message):
            accountant.measure(measurement=measurement)

    def test_measure(self):
        """PrivacyAccountant.measure answers measurement correctly."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        mock_measurement = create_mock_measurement(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_measure=self.output_measure,
            privacy_function_implemented=True,
            privacy_function_return_value=self.budget_quarters[3],
            return_value=np.int64(2),
        )
        actual_answer = accountant.measure(measurement=mock_measurement)
        self.assertEqual(actual_answer, np.int64(2))
        self.assertEqual(accountant.privacy_budget, self.budget_quarters[1])

    def test_measure_consumes_all_privacy_budget(self):
        """Test that measurements are allowed to consume the entire budget."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        mock_measurement = create_mock_measurement(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_measure=self.output_measure,
            privacy_function_implemented=True,
            privacy_function_return_value=accountant.privacy_budget,
            return_value=np.int64(2),
        )
        actual_answer = accountant.measure(measurement=mock_measurement)
        self.assertEqual(actual_answer, np.int64(2))
        self.assertEqual(accountant.privacy_budget, self.budget_quarters[0])

    @parameterized.expand(
        [
            (
                "Given d_out does not satisfy the stability relation of given",
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                PureDP(),
                8,
                False,
                1,
            ),
            (
                (
                    "Transformation's input domain does not match PrivacyAccountant's"
                    " input domain"
                ),
                NumpyStringDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                PureDP(),
                8,
            ),
            (
                (
                    "Transformation's input metric does not match PrivacyAccountant's"
                    " input metric"
                ),
                NumpyIntegerDomain(),
                SymmetricDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                PureDP(),
                8,
            ),
            (
                "Splitting transformation's output domain must be ListDomain.",
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                NumpyIntegerDomain(),
                SumOf(AbsoluteDifference()),
                PureDP(),
                8,
            ),
            (
                "Splitting transformation's output domain must specify list length.",
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain()),
                SumOf(AbsoluteDifference()),
                PureDP(),
                8,
            ),
            (
                (
                    "Splitting transformation's output metric must be"
                    f" {SumOf} for output measure PureDP"
                ),
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                RootSumOfSquared(AbsoluteDifference()),
                PureDP(),
                8,
            ),
            (
                (
                    "Splitting transformation's output metric must be"
                    f" {RootSumOfSquared} for output measure RhoZCDP"
                ),
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                RhoZCDP(),
                8,
            ),
            (
                r"PrivacyAccountant's remaining privacy budget is 10, which is"
                r" insufficient for this operation that requires privacy loss 11.",
                NumpyIntegerDomain(),
                AbsoluteDifference(),
                ListDomain(NumpyIntegerDomain(), length=2),
                SumOf(AbsoluteDifference()),
                PureDP(),
                11,
            ),
        ]
    )
    def test_split_invalid_arguments(
        self,
        error_message: str,
        input_domain: Domain,
        input_metric: Metric,
        output_domain: Domain,
        output_metric: Metric,
        output_measure: Union[PureDP, RhoZCDP],
        privacy_budget: ExactNumberInput,
        stability_function_implemented: bool = True,
        d_out: Optional[Any] = None,
    ):
        """PrivacyAccountant.split raises errors appropriately."""
        accountant = PrivacyAccountant.launch(
            measurement=SequentialComposition(
                input_domain=NumpyIntegerDomain(),
                input_metric=AbsoluteDifference(),
                output_measure=output_measure,
                d_in=1,
                privacy_budget=10,
            ),
            data=np.int64(10),
        )
        splitting_transformation = create_mock_transformation(
            input_domain=input_domain,
            input_metric=input_metric,
            output_domain=output_domain,
            output_metric=output_metric,
            stability_function_implemented=stability_function_implemented,
            stability_function_return_value=5,
            stability_relation_return_value=False,
            return_value=[np.int64(2), np.int64(3)],
        )
        with self.assertRaisesRegex(
            (ValueError, UnsupportedDomainError), error_message
        ):
            accountant.split(
                splitting_transformation=splitting_transformation,
                d_out=d_out,
                privacy_budget=privacy_budget,
            )

    def test_split(self):
        """PrivacyAccountant.split works correctly."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=ListDomain(element_domain=NumpyIntegerDomain(), length=2),
            output_metric=self.splitting_output_metric,
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=[np.int64(2), np.int64(3)],
        )
        split_budget = self.budget_quarters[1]
        child_accountants = accountant.split(
            splitting_transformation=transformation, privacy_budget=split_budget
        )
        self.assertEqual(accountant.privacy_budget, self.budget_quarters[3])
        self.assertEqual(accountant.state, PrivacyAccountantState.WAITING_FOR_CHILDREN)
        self.assertEqual(len(child_accountants), 2)
        for index, child in enumerate(child_accountants):
            self.assertEqual(
                child.input_domain, transformation.output_domain.element_domain
            )
            self.assertEqual(
                child.input_metric, transformation.output_metric.inner_metric
            )
            self.assertEqual(child.output_measure, accountant.output_measure)
            self.assertEqual(child.d_in, 10)
            self.assertEqual(child.privacy_budget, split_budget)
            expected_state = (
                PrivacyAccountantState.ACTIVE
                if index == 0
                else PrivacyAccountantState.WAITING_FOR_SIBLING
            )
            self.assertEqual(child.state, expected_state)

    def test_queue_transformation_on_active_accountant(self):
        """queue_transformation runs immediately on active accountant"""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=NumpyIntegerDomain(),
            output_metric=AbsoluteDifference(),
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=np.int64(2),
        )
        accountant.queue_transformation(transformation=transformation)
        self.assertEqual(accountant.input_domain, NumpyIntegerDomain())
        self.assertEqual(accountant.input_metric, AbsoluteDifference())
        self.assertEqual(accountant.d_in, 10)
        # pylint: disable=protected-access
        self.assertIsNotNone(accountant._queryable)
        self.assertEqual(accountant._queryable._data, np.int64(2))  # type: ignore
        self.assertIsNone(accountant._pending_transformation)
        # pylint: enable=protected-access

    def test_queue_transformation_on_inactive_accountant(self):
        """queue_transformation queues transformations on inactive account"""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        split_transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=ListDomain(element_domain=NumpyIntegerDomain(), length=2),
            output_metric=self.splitting_output_metric,
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=[np.int64(2), np.int64(3)],
        )
        split_budget = self.budget_quarters[1]
        child_accountants = accountant.split(
            splitting_transformation=split_transformation, privacy_budget=split_budget
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=NumpyIntegerDomain(),
            output_metric=AbsoluteDifference(),
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=np.int64(2),
        )
        accountant.queue_transformation(transformation=transformation)
        # values should reflect the pending transformation
        self.assertEqual(accountant.input_domain, NumpyIntegerDomain())
        self.assertEqual(accountant.input_metric, AbsoluteDifference())
        self.assertEqual(accountant.d_in, 10)
        self.assertIsNotNone(
            accountant._pending_transformation  # pylint: disable=protected-access
        )

        for c in child_accountants:
            c.retire()

        # Once the accountant is active again, the transformation should
        # have been run
        self.assertEqual(accountant.state, PrivacyAccountantState.ACTIVE)
        self.assertEqual(accountant.input_domain, NumpyIntegerDomain())
        self.assertEqual(accountant.input_metric, AbsoluteDifference())
        self.assertEqual(accountant.d_in, 10)
        # pylint: disable=protected-access
        self.assertIsNotNone(accountant._queryable)
        self.assertEqual(accountant._queryable._data, np.int64(2))  # type: ignore
        self.assertIsNone(accountant._pending_transformation)
        # pylint: enable=protected-access

    @parameterized.expand(
        [
            (
                (
                    "Transformation's input domain does not match PrivacyAccountant's"
                    " input domain"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {"A": SparkIntegerColumnDescriptor()}
                    ),
                    input_metric=SymmetricDifference(),
                    stability_function_implemented=True,
                ),
            ),
            (
                (
                    "Transformation's input metric does not match PrivacyAccountant's"
                    " input metric"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {
                            "A": SparkIntegerColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                        }
                    ),
                    input_metric=HammingDistance(),
                    stability_function_implemented=True,
                ),
            ),
        ]
    )
    def test_queue_transformation_invalid_arguments(
        self,
        error_message: str,
        transformation: Transformation,
        d_out: Optional[Any] = None,
    ) -> None:
        """Test queue_transformation with an invalid transformation."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        split_transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=ListDomain(element_domain=NumpyIntegerDomain(), length=2),
            output_metric=self.splitting_output_metric,
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=[np.int64(2), np.int64(3)],
        )
        split_budget = self.budget_quarters[1]
        accountant.split(
            splitting_transformation=split_transformation, privacy_budget=split_budget
        )
        self.assertEqual(accountant.state, PrivacyAccountantState.WAITING_FOR_CHILDREN)
        with self.assertRaisesRegex(ValueError, error_message):
            accountant.queue_transformation(transformation=transformation, d_out=d_out)

    @parameterized.expand(
        [
            (
                (
                    "Transformation's input domain does not match the output domain"
                    " of the last transformation"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {"A": SparkIntegerColumnDescriptor()}
                    ),
                    input_metric=SymmetricDifference(),
                    stability_function_implemented=True,
                ),
            ),
            (
                (
                    "Transformation's input metric does not match the output metric"
                    " of the last transformation"
                ),
                create_mock_transformation(
                    input_domain=SparkDataFrameDomain(
                        {
                            "A": SparkIntegerColumnDescriptor(),
                            "B": SparkStringColumnDescriptor(),
                        }
                    ),
                    input_metric=HammingDistance(),
                    stability_function_implemented=True,
                ),
            ),
        ]
    )
    def test_queue_invalid_transformation_with_transform_in_queue(
        self,
        error_message: str,
        transformation: Transformation,
        d_out: Optional[Any] = None,
    ):
        """Test queue_transformation with invalid arguments and a pending transform."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        split_transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=ListDomain(element_domain=NumpyIntegerDomain(), length=2),
            output_metric=self.splitting_output_metric,
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=[np.int64(2), np.int64(3)],
        )
        split_budget = self.budget_quarters[1]
        accountant.split(
            splitting_transformation=split_transformation, privacy_budget=split_budget
        )
        self.assertEqual(accountant.state, PrivacyAccountantState.WAITING_FOR_CHILDREN)

        identity_transformation = create_mock_transformation(
            input_domain=accountant.input_domain,
            input_metric=accountant.input_metric,
            output_domain=accountant.input_domain,
            output_metric=accountant.input_metric,
            stability_function_implemented=True,
        )
        accountant.queue_transformation(identity_transformation)
        self.assertIsNotNone(
            accountant._pending_transformation  # pylint: disable=protected-access
        )
        with self.assertRaisesRegex(ValueError, error_message):
            accountant.queue_transformation(transformation=transformation, d_out=d_out)

    def test_force_activate_raises_error_on_invalid_states(self):
        """PrivacyAccountant.force_activate raises error appropriately."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        accountant._state = (  # pylint: disable=protected-access
            PrivacyAccountantState.RETIRED
        )
        with self.assertRaisesRegex(
            RuntimeError, "Can not activate RETIRED PrivacyAccountant"
        ):
            accountant.force_activate()

    def test_force_activate_active(self):
        """PrivacyAccountant.force_activate works when ACTIVE."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        accountant.force_activate()
        self.assertEqual(accountant.state, PrivacyAccountantState.ACTIVE)

    def test_force_activate_waiting_for_siblings(self):
        """PrivacyAccountant.force_activate works when WAITING_FOR_SIBLINGS."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        transformation = create_mock_transformation(
            input_domain=SparkDataFrameDomain(
                {
                    "A": SparkIntegerColumnDescriptor(),
                    "B": SparkStringColumnDescriptor(),
                }
            ),
            input_metric=SymmetricDifference(),
            output_domain=ListDomain(element_domain=NumpyIntegerDomain(), length=4),
            output_metric=self.splitting_output_metric,
            stability_function_implemented=True,
            stability_function_return_value=10,
            return_value=[np.int64(2), np.int64(3), np.int64(4), np.int64(5)],
        )
        child_accountants = accountant.split(
            splitting_transformation=transformation,
            privacy_budget=self.budget_quarters[1],
        )

        child_accountants[2].force_activate()

        for i in range(0, 2):
            self.assertEqual(child_accountants[i].state, PrivacyAccountantState.RETIRED)

        self.assertEqual(child_accountants[2].state, PrivacyAccountantState.ACTIVE)
        self.assertEqual(
            child_accountants[3].state, PrivacyAccountantState.WAITING_FOR_SIBLING
        )
        self.assertEqual(accountant.state, PrivacyAccountantState.WAITING_FOR_CHILDREN)

    def test_retire_raises_error_appropriately(self):
        """PrivacyAccountant.retire raises error appropriately."""
        # pylint: disable=protected-access
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        accountant._state = PrivacyAccountantState.WAITING_FOR_CHILDREN
        with self.assertRaisesRegex(
            RuntimeError,
            "Can not retire PrivacyAccountant in WAITING_FOR_CHILDREN state",
        ):
            accountant.retire()

    def test_retire_raises_warning_appropriately(self):
        """PrivacyAccountant.retire raises warning appropriately."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        children = accountant.split(
            splitting_transformation=create_mock_transformation(
                input_domain=self.measurement.input_domain,
                input_metric=self.measurement.input_metric,
                output_domain=ListDomain(NumpyIntegerDomain(), length=4),
                output_metric=self.splitting_output_metric,
                return_value=[np.int64(0) for _ in range(4)],
                stability_function_implemented=True,
            ),
            privacy_budget=self.budget_quarters[1],
        )
        with self.assertWarnsRegex(
            RuntimeWarning,
            (
                "Retiring an unused PrivacyAccountant that is"
                " PrivacyAccountantState.WAITING_FOR_SIBLING"
            ),
        ):
            children[1].retire()

    def test_retire_accountant_waiting_for_sibling(self):
        """Retiring an unused sibling retires all prior siblings and activates next."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        children = accountant.split(
            splitting_transformation=create_mock_transformation(
                input_domain=self.measurement.input_domain,
                input_metric=self.measurement.input_metric,
                output_domain=ListDomain(NumpyIntegerDomain(), length=4),
                output_metric=self.splitting_output_metric,
                return_value=[np.int64(0) for _ in range(4)],
                stability_function_implemented=True,
            ),
            privacy_budget=self.budget_quarters[1],
        )
        assert len(children) == 4
        assert children[0].state == PrivacyAccountantState.ACTIVE
        assert (
            children[1].state
            == children[2].state
            == PrivacyAccountantState.WAITING_FOR_SIBLING
        )
        children[2].retire()
        self.assertTrue(
            all(
                child.state == PrivacyAccountantState.RETIRED for child in children[:-1]
            )
        )
        self.assertEqual(children[3].state, PrivacyAccountantState.ACTIVE)

    def test_retiring_all_children_activates_parent(self):
        """PrivacyAccount transitions from WAITING_FOR_CHILDREN to ACTIVE correctly."""
        accountant = PrivacyAccountant.launch(
            measurement=self.measurement, data=self.data
        )
        children = accountant.split(
            splitting_transformation=create_mock_transformation(
                input_domain=self.measurement.input_domain,
                input_metric=self.measurement.input_metric,
                output_domain=ListDomain(NumpyIntegerDomain(), length=2),
                output_metric=self.splitting_output_metric,
                return_value=[np.int64(0) for _ in range(2)],
                stability_function_implemented=True,
            ),
            privacy_budget=self.budget_quarters[1],
        )
        assert accountant.state == PrivacyAccountantState.WAITING_FOR_CHILDREN
        for child in children:
            child.retire()

        self.assertEqual(accountant.state, PrivacyAccountantState.ACTIVE)


@parameterized_class(
    [
        {"output_measure": PureDP(), "privacy_budget": 6},
        {"output_measure": ApproxDP(), "privacy_budget": (6, sp.Rational("0.3"))},
        {"output_measure": RhoZCDP(), "privacy_budget": 6},
    ]
)
class TestDecorateQueryable(TestCase):
    """Tests for DecorateQueryable."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    privacy_budget: Union[int, Tuple[int, sp.Rational]]
    """The privacy budget to use during the tests."""

    def setUp(self):
        """Set up class."""
        self.mock_queryable = create_mock_queryable()
        self.measurement = DecorateQueryable(
            measurement=create_mock_measurement(
                input_domain=NumpyIntegerDomain(),
                input_metric=AbsoluteDifference(),
                output_measure=self.output_measure,
                is_interactive=True,
                privacy_function_implemented=True,
                privacy_function_return_value=self.privacy_budget,
                return_value=self.mock_queryable,
            ),
            preprocess_query=lambda x: x,
            postprocess_answer=lambda x: x,
        )

    @parameterized.expand(get_all_props(DecorateQueryable))
    def test_property_immutability(self, prop_name: str):
        """Tests that given property is immutable."""
        assert_property_immutability(self.measurement, prop_name)

    def test_properties(self):
        """SequentialComposition's properties have the expected values."""
        self.assertEqual(self.measurement.input_domain, NumpyIntegerDomain())
        self.assertEqual(self.measurement.input_metric, AbsoluteDifference())
        self.assertEqual(self.measurement.output_measure, self.output_measure)
        self.assertEqual(self.measurement.is_interactive, True)

    def test_privacy_function(self):
        """SequentialComposition's privacy function is correct."""
        self.assertEqual(self.measurement.privacy_function(1), self.privacy_budget)

    def test_correctness(self):
        """SequentialComposition returns the expected Queryable object."""
        # pylint: disable=protected-access
        actual = self.measurement(np.int64(10))
        self.assertIsInstance(actual, DecoratedQueryable)
        self.assertEqual(actual._preprocess_query, self.measurement.preprocess_query)
        self.assertEqual(
            actual._postprocess_answer, self.measurement.postprocess_answer
        )
        self.assertEqual(actual._queryable, self.mock_queryable)


@parameterized_class(
    [
        {"output_measure": PureDP(), "privacy_budget": 6},
        {"output_measure": ApproxDP(), "privacy_budget": (6, sp.Rational("0.3"))},
        {"output_measure": RhoZCDP(), "privacy_budget": 6},
    ]
)
class TestDecoratedQueryable(TestCase):
    """Tests for DecoratedQueryable."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    privacy_budget: Union[int, Tuple[int, sp.Rational]]
    """The privacy budget to use during the tests."""

    def test_correctness(self):
        """Set up class."""
        mock_queryable = MagicMock()
        mock_queryable.return_value = 1
        decorated_queryable = DecorateQueryable(
            measurement=create_mock_measurement(
                input_domain=NumpyIntegerDomain(),
                input_metric=AbsoluteDifference(),
                output_measure=self.output_measure,
                is_interactive=True,
                privacy_function_implemented=True,
                privacy_function_return_value=self.privacy_budget,
                return_value=mock_queryable,
            ),
            preprocess_query=lambda x: 2 * x,
            postprocess_answer=lambda x: x + 1,
        )(np.int64(1))
        actual = decorated_queryable(2)
        mock_queryable.assert_called_once_with(4)
        self.assertEqual(actual, 2)


@parameterized_class(
    [
        {
            "output_measure": PureDP(),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
        {
            "output_measure": ApproxDP(),
            "budget_quarters": [
                (0, 0),
                (sp.Rational("2.5"), sp.Rational("0.1")),
                (5, sp.Rational("0.2")),
                (sp.Rational("7.5"), sp.Rational("0.3")),
                (10, sp.Rational("0.4")),
            ],
        },
        {
            "output_measure": RhoZCDP(),
            "budget_quarters": [0, sp.Rational("2.5"), 5, sp.Rational("7.5"), 10],
        },
    ]
)
class TestCreateAdaptiveComposition(TestCase):
    """Tests for :func:`~.create_adaptive_composition`."""

    output_measure: Union[PureDP, ApproxDP, RhoZCDP]
    """The output measure to use during the tests."""

    budget_quarters: List
    """Zero, one quarter, one half, three quarters, and all of the privacy budget."""

    def setUp(self):
        """Test setup."""
        self.privacy_budget = self.budget_quarters[4]
        self.queryable = create_adaptive_composition(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            d_in=1,
            privacy_budget=self.privacy_budget,
            output_measure=self.output_measure,
        )(np.int64(10))

    def test_create_adaptive_composition(self):
        """:func:`~.create_adaptive_composition` works as expected."""
        adaptive_composition = create_adaptive_composition(
            input_domain=NumpyIntegerDomain(),
            input_metric=AbsoluteDifference(),
            d_in=1,
            privacy_budget=self.privacy_budget,
            output_measure=self.output_measure,
        )
        self.assertIsInstance(adaptive_composition, DecorateQueryable)
        self.assertIsInstance(adaptive_composition.measurement, SequentialComposition)
        self.assertEqual(
            adaptive_composition.measurement.input_domain, NumpyIntegerDomain()
        )
        self.assertEqual(
            adaptive_composition.measurement.input_metric, AbsoluteDifference()
        )
        self.assertEqual(adaptive_composition.measurement.d_in, 1)  # type: ignore
        # pylint: disable=line-too-long
        self.assertEqual(
            adaptive_composition.measurement.privacy_budget, self.privacy_budget  # type: ignore
        )
        # pylint: enable=line-too-long
        self.assertEqual(
            adaptive_composition.measurement.output_measure, self.output_measure
        )
        self.assertIsInstance(adaptive_composition(np.int64(10)), DecoratedQueryable)

    def test_correctness(self):
        """Queryable works as expected."""
        query = MeasurementQuery(
            create_mock_measurement(
                output_measure=self.output_measure,
                privacy_function_implemented=True,
                privacy_function_return_value=self.budget_quarters[1],
                return_value=np.int64(22),
            )
        )
        answer = self.queryable(query)
        self.assertEqual(answer, np.int64(22))

    def test_interactive_measurement(self):
        """Can not answer interactive measurement queries."""
        query = MeasurementQuery(create_mock_measurement(is_interactive=True))
        with self.assertRaisesRegex(
            ValueError, "Cannot answer interactive measurement query"
        ):
            self.queryable(query)

    def test_insufficient_budget(self):
        """Raises error on insufficient budget."""
        query1 = MeasurementQuery(
            create_mock_measurement(
                output_measure=self.output_measure,
                privacy_function_implemented=True,
                privacy_function_return_value=self.budget_quarters[2],
            )
        )
        self.queryable(query1)
        query2 = MeasurementQuery(
            create_mock_measurement(
                output_measure=self.output_measure,
                privacy_function_implemented=True,
                privacy_function_return_value=self.budget_quarters[3],
            )
        )
        with self.assertRaisesRegex(
            ValueError, "Cannot answer query without exceeding available privacy budget"
        ):
            self.queryable(query2)



"""Unit tests for :mod:`tmlt.core.utils.join`."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

import re
from datetime import date, datetime
from typing import Any, Dict, List, Optional, Tuple, Type
from unittest import TestCase

from parameterized import parameterized
from pyspark.sql import DataFrame

from tmlt.core.domains.numpy_domains import NumpyIntegerDomain
from tmlt.core.domains.spark_domains import (
    SparkColumnDescriptor,
    SparkDataFrameDomain,
    SparkDateColumnDescriptor,
    SparkFloatColumnDescriptor,
    SparkIntegerColumnDescriptor,
    SparkStringColumnDescriptor,
    SparkTimestampColumnDescriptor,
)
from tmlt.core.utils.join import (
    columns_after_join,
    domain_after_join,
    join,
    natural_join_columns,
)
from tmlt.core.utils.testing import PySparkTest


class TestNaturalJoinColumns(TestCase):
    """Tests for :func:`tmlt.core.utils.join.natural_join_columns`."""

    @parameterized.expand(
        [
            (["a", "b", "c"], ["a", "b", "c"], ["a", "b", "c"]),
            (["a", "b", "c"], ["d", "c", "b", "a"], ["a", "b", "c"]),
            (["a", "b", "c"], ["b", "d"], ["b"]),
            (["a", "b", "c"], ["d", "e", "f"], []),
        ]
    )
    def test_correctness(
        self,
        left_columns: List[str],
        right_columns: List[str],
        expected_columns: List[str],
    ):
        """Test that the output is correct."""
        self.assertEqual(
            natural_join_columns(left_columns, right_columns), expected_columns
        )


COLUMNS_AFTER_JOIN_CORRECTNESS_TEST_CASES = [
    (
        ["a", "b", "c"],
        ["a", "b", "c"],
        None,
        {"a": ("a", "a"), "b": ("b", "b"), "c": ("c", "c")},
    ),
    (
        ["a", "b", "c"],
        ["d", "c", "b", "a"],
        None,
        {"a": ("a", "a"), "b": ("b", "b"), "c": ("c", "c"), "d": (None, "d")},
    ),
    (
        ["a", "b", "c"],
        ["b", "d"],
        None,
        {"b": ("b", "b"), "a": ("a", None), "c": ("c", None), "d": (None, "d")},
    ),
    (
        ["a", "b", "c"],
        ["a", "b", "c"],
        ["a"],
        {
            "a": ("a", "a"),
            "b_left": ("b", None),
            "c_left": ("c", None),
            "b_right": (None, "b"),
            "c_right": (None, "c"),
        },
    ),
    (
        ["a", "b", "c"],
        ["d", "c", "a"],
        ["a"],
        {
            "a": ("a", "a"),
            "b": ("b", None),
            "c_left": ("c", None),
            "d": (None, "d"),
            "c_right": (None, "c"),
        },
    ),
    (
        ["a_left", "b_left", "c_right"],
        ["a_right", "b_left", "c_right"],
        ["b_left"],
        {
            "b_left": ("b_left", "b_left"),
            "a_left": ("a_left", None),
            "c_right_left": ("c_right", None),
            "a_right": (None, "a_right"),
            "c_right_right": (None, "c_right"),
        },
    ),
]

COLUMNS_AFTER_JOIN_ERROR_TEST_CASES: Any = [
    (["a", "b", "c"], ["a", "b", "c"], [], "Join must involve at least one column."),
    (["a", "b", "c"], ["d"], None, "Join must involve at least one column."),
    (["a", "b", "c"], ["a", "d"], ["d"], "Join column 'd' not in the left table."),
    (["a", "b", "c"], ["a", "d"], ["b"], "Join column 'b' not in the right table."),
    (["a", "a", "b"], ["b", "c"], None, "Left columns contain duplicates."),
    (["a", "b", "c"], ["b", "b", "c"], None, "Right columns contain duplicates."),
    (
        ["a", "b", "c"],
        ["a", "c"],
        ["a", "a"],
        "Join columns (`on`) contain duplicates.",
    ),
    (
        ["a", "b"],
        ["a", "b", "b_right"],
        ["a"],
        "Name collision, 'b_right' would appear more than once in the output.",
    ),
    (
        ["a", "b"],
        ["a", "b_right", "b"],
        ["a"],
        "Name collision, 'b_right' would appear more than once in the output.",
    ),
    (
        ["a", "b", "b_left"],
        ["a", "b"],
        ["a"],
        "Name collision, 'b_left' would appear more than once in the output.",
    ),
    (
        ["a", "b_left", "b"],
        ["a", "b"],
        ["a"],
        "Name collision, 'b_left' would appear more than once in the output.",
    ),
]

DOMAIN_AFTER_JOIN_ERROR_TEST_CASES = [
    (
        {"how": "invalid"},
        (
            "Join type (`how`) must be one of 'left', 'right', 'inner', or "
            "'outer', not 'invalid'"
        ),
    ),
    (
        {
            "left_domain": SparkDataFrameDomain(
                {"column": SparkIntegerColumnDescriptor()}
            )
        },
        "'column' has different data types in left (LongType) and"
        " right (StringType) domains.",
    ),
    (
        {
            "left_domain": SparkDataFrameDomain(
                {"column": SparkIntegerColumnDescriptor(size=32)}
            ),
            "right_domain": SparkDataFrameDomain(
                {"column": SparkIntegerColumnDescriptor(size=64)}
            ),
        },
        "'column' has different data types in left (IntegerType) and right "
        "(LongType) domains.",
    ),
    (
        {"left_domain": NumpyIntegerDomain()},
        "Left join input domain must be a SparkDataFrameDomain.",
    ),
    (
        {"right_domain": NumpyIntegerDomain()},
        "Right join input domain must be a SparkDataFrameDomain.",
    ),
]


class TestColumnsAfterJoin(TestCase):
    """Tests for :func:`tmlt.core.utils.join.columns_after_join`."""

    @parameterized.expand(COLUMNS_AFTER_JOIN_CORRECTNESS_TEST_CASES)
    def test_correctness(
        self,
        left_columns: List[str],
        right_columns: List[str],
        on: Optional[List[str]],
        expected_columns: Dict[str, Tuple[Optional[str], Optional[str]]],
    ):
        """Test that the output is correct."""
        self.assertEqual(
            columns_after_join(left_columns, right_columns, on), expected_columns
        )

    @parameterized.expand(COLUMNS_AFTER_JOIN_ERROR_TEST_CASES)
    def test_validation(
        self,
        left_columns: List[str],
        right_columns: List[str],
        on: Optional[List[str]],
        message: str,
    ):
        """Test that the expected exception is raised when the input is invalid."""
        with self.assertRaisesRegex(ValueError, re.escape(message)):
            columns_after_join(left_columns, right_columns, on)


class TestDomainAfterJoin(TestCase):
    """Tests for :func:`tmlt.core.utils.join.domain_after_join`."""

    @parameterized.expand(COLUMNS_AFTER_JOIN_CORRECTNESS_TEST_CASES)
    def test_columns_after_join_correctness(
        self,
        left_columns: List[str],
        right_columns: List[str],
        on: Optional[List[str]],
        expected_columns: Dict[str, Tuple[Optional[str], Optional[str]]],
    ):
        """Test that domain_after_join preserves behavior of columns_after_join."""
        left_domain = SparkDataFrameDomain(
            {column: SparkStringColumnDescriptor() for column in left_columns}
        )
        right_domain = SparkDataFrameDomain(
            {column: SparkStringColumnDescriptor() for column in right_columns}
        )
        expected_domain = SparkDataFrameDomain(
            {column: SparkStringColumnDescriptor() for column in expected_columns}
        )
        self.assertEqual(
            domain_after_join(left_domain, right_domain, on), expected_domain
        )

    @parameterized.expand(
        [how, left_allow_nan, left_allow_inf, right_allow_nan, right_allow_inf]
        for how in ["inner", "left", "right", "outer"]
        for left_allow_nan in [True, False]
        for left_allow_inf in [True, False]
        for right_allow_nan in [True, False]
        for right_allow_inf in [True, False]
    )
    def test_floating_point_special_values(
        self,
        how: str,
        left_allow_nan: bool,
        left_allow_inf: bool,
        right_allow_nan: bool,
        right_allow_inf: bool,
    ):
        """Test that special values in floating point columns are handled correctly."""
        left_domain = SparkDataFrameDomain(
            {
                "joined_on": SparkFloatColumnDescriptor(
                    allow_nan=left_allow_nan, allow_inf=left_allow_inf
                )
            }
        )
        right_domain = SparkDataFrameDomain(
            {
                "joined_on": SparkFloatColumnDescriptor(
                    allow_nan=right_allow_nan, allow_inf=right_allow_inf
                ),
                "not_joined_on": SparkFloatColumnDescriptor(
                    allow_nan=left_allow_nan, allow_inf=left_allow_inf
                ),
            }
        )
        if how == "left":
            allow_inf = left_allow_inf
            allow_nan = left_allow_nan
        elif how == "right":
            allow_inf = right_allow_inf
            allow_nan = right_allow_nan
        elif how == "inner":
            allow_inf = left_allow_inf and right_allow_inf
            allow_nan = left_allow_nan and right_allow_nan
        else:
            allow_inf = left_allow_inf or right_allow_inf
            allow_nan = left_allow_nan or right_allow_nan
        expected_domain = SparkDataFrameDomain(
            {
                "joined_on": SparkFloatColumnDescriptor(
                    allow_nan=allow_nan, allow_inf=allow_inf
                ),
                "not_joined_on": SparkFloatColumnDescriptor(
                    allow_nan=left_allow_nan,
                    allow_inf=left_allow_inf,
                    allow_null=how in ["left", "outer"],
                ),
            }
        )
        self.assertEqual(
            domain_after_join(left_domain, right_domain, how=how), expected_domain
        )

    @parameterized.expand(
        [how, left_allow_null, right_allow_null, nulls_are_equal, descriptor_class]
        for how in ["inner", "left", "right", "outer"]
        for left_allow_null in [True, False]
        for right_allow_null in [True, False]
        for nulls_are_equal in [True, False]
        for descriptor_class in [
            SparkStringColumnDescriptor,
            SparkFloatColumnDescriptor,
            SparkIntegerColumnDescriptor,
            SparkDateColumnDescriptor,
            SparkTimestampColumnDescriptor,
        ]
    )
    def test_null_values(
        self,
        how: str,
        left_allow_null: bool,
        right_allow_null: bool,
        nulls_are_equal: bool,
        descriptor_class: Type[SparkColumnDescriptor],
    ):
        """Test that null values are handled correctly."""
        left_domain = SparkDataFrameDomain(
            {
                "joined_on": descriptor_class(  # type: ignore
                    allow_null=left_allow_null
                ),
                "not_joined_on": descriptor_class(  # type: ignore
                    allow_null=left_allow_null
                ),
            }
        )
        right_domain = SparkDataFrameDomain(
            {"joined_on": descriptor_class(allow_null=right_allow_null)}  # type: ignore
        )
        if how == "left":
            allow_null = left_allow_null
        elif how == "right":
            allow_null = right_allow_null
        elif how == "inner":
            allow_null = nulls_are_equal and left_allow_null and right_allow_null
        else:
            allow_null = left_allow_null or right_allow_null
        expected_domain = SparkDataFrameDomain(
            {
                "joined_on": descriptor_class(allow_null=allow_null),  # type: ignore
                "not_joined_on": descriptor_class(  # type: ignore
                    allow_null=(left_allow_null or how in ["right", "outer"])
                ),
            }
        )
        self.assertEqual(
            domain_after_join(
                left_domain, right_domain, how=how, nulls_are_equal=nulls_are_equal
            ),
            expected_domain,
        )

    @parameterized.expand(COLUMNS_AFTER_JOIN_ERROR_TEST_CASES)
    def test_columns_after_join_validation(
        self,
        left_columns: List[str],
        right_columns: List[str],
        on: Optional[List[str]],
        message: str,
    ):
        """Test that domain_after_join preserves validation of columns_after_join."""
        if len(left_columns) != len(set(left_columns)):
            return  # SparkDataFrameDomains can't have duplicates
        if len(right_columns) != len(set(right_columns)):
            return
        left_domain = SparkDataFrameDomain(
            {column: SparkStringColumnDescriptor() for column in left_columns}
        )
        right_domain = SparkDataFrameDomain(
            {column: SparkStringColumnDescriptor() for column in right_columns}
        )
        with self.assertRaisesRegex(ValueError, re.escape(message)):
            domain_after_join(left_domain, right_domain, on)

    @parameterized.expand(DOMAIN_AFTER_JOIN_ERROR_TEST_CASES)
    def test_validation(self, args_updates: Dict[str, Any], message: str):
        """Test that invalid inputs raise an error."""
        args = {
            "left_domain": SparkDataFrameDomain(
                {"column": SparkStringColumnDescriptor()}
            ),
            "right_domain": SparkDataFrameDomain(
                {"column": SparkStringColumnDescriptor()}
            ),
            "on": ["column"],
            "how": "inner",
            "nulls_are_equal": True,
        }
        args.update(args_updates)

        with self.assertRaisesRegex(Exception, re.escape(message)):
            domain_after_join(**args)  # type: ignore


class TestJoin(PySparkTest):
    """Tests for :func:`tmlt.core.utils.join.join`."""

    @parameterized.expand(
        [
            (  # Basic inner join with equal nulls
                {
                    "A": [1, 1, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [1, 1, 3, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "inner",
                True,
                {
                    "A": [1, 1, 1, 1, 3, None],
                    "B": ["a", "a", "b", "b", None, None],
                    "C": [1.0, 1.0, 2.0, 2.0, 3.0, 4.0],
                    "D": ["a", "b", "a", "b", "c", "d"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
            ),
            (  # Basic inner join without equal nulls
                {
                    "A": [1, 1, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [1, 1, 3, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "inner",
                False,
                {
                    "A": [1, 1, 1, 1, 3],
                    "B": ["a", "a", "b", "b", None],
                    "C": [1.0, 1.0, 2.0, 2.0, 3.0],
                    "D": ["a", "b", "a", "b", "c"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
            ),
            (  # Basic left join without equal nulls
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [None, 1, 3, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "left",
                False,
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                    "D": ["b", None, "c", None],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                        "D": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # Basic right join with equal nulls
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [None, 1, 4, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "right",
                True,
                {
                    "A": [None, None, 1, 4, None, None],
                    "B": ["b", None, "a", None, "b", None],
                    "C": [2.0, 4.0, 1.0, None, 2.0, 4.0],
                    "D": ["a", "a", "b", "c", "d", "d"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
            ),
            (  # Basic left join with equal nulls
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [None, 1, 4, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "left",
                True,
                {
                    "A": [1, None, None, 3, None, None],
                    "B": ["a", "b", "b", None, None, None],
                    "C": [1.0, 2.0, 2.0, 3.0, 4.0, 4.0],
                    "D": ["b", "a", "d", None, "a", "d"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                        "D": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # Basic outer join with equal nulls
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [None, 1, 4, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "outer",
                True,
                {
                    "A": [1, None, None, 3, None, None, 4],
                    "B": ["a", "b", "b", None, None, None, None],
                    "C": [1.0, 2.0, 2.0, 3.0, 4.0, 4.0, None],
                    "D": ["b", "a", "d", None, "a", "d", "c"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # Basic outer join without equal nulls
                {
                    "A": [1, None, 3, None],
                    "B": ["a", "b", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [None, 1, 4, None], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "outer",
                False,
                {
                    "A": [1, None, 3, None, None, 4, None],
                    "B": ["a", "b", None, None, None, None, None],
                    "C": [1.0, 2.0, 3.0, 4.0, None, None, None],
                    "D": ["b", None, None, None, "a", "c", "d"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=True),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # Outer join, join columns don't have nulls
                {
                    "A": [1, 2, 3, 4],
                    "B": ["a", "b", "c", "d"],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "B": SparkStringColumnDescriptor(allow_null=False),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [3, 4, 5, 6], "D": ["a", "b", "c", "d"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "D": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "outer",
                True,
                {
                    "A": [1, 2, 3, 4, 5, 6],
                    "B": ["a", "b", "c", "d", None, None],
                    "C": [1.0, 2.0, 3.0, 4.0, None, None],
                    "D": [None, None, "a", "b", "c", "d"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=True),
                        "D": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # Float special values outer join
                {"A": [1.0, float("inf"), float("-inf"), float("nan"), None]},
                SparkDataFrameDomain(
                    {
                        "A": SparkFloatColumnDescriptor(
                            allow_null=True, allow_nan=True, allow_inf=True
                        )
                    }
                ),
                {"A": [1.0, float("inf"), float("-inf"), float("nan"), None]},
                SparkDataFrameDomain(
                    {
                        "A": SparkFloatColumnDescriptor(
                            allow_null=True, allow_nan=True, allow_inf=True
                        )
                    }
                ),
                ["A"],
                "outer",
                True,
                {"A": [1.0, float("inf"), float("-inf"), float("nan"), None]},
                SparkDataFrameDomain(
                    {
                        "A": SparkFloatColumnDescriptor(
                            allow_null=True, allow_nan=True, allow_inf=True
                        )
                    }
                ),
            ),
            (  # Date and timestamp outer join
                {
                    "timestamp": [
                        datetime(2020, 1, 1),
                        datetime(2020, 1, 2),
                        datetime(2020, 1, 3),
                        datetime(2020, 1, 4),
                    ],
                    "date": [
                        date(2020, 1, 1),
                        date(2020, 1, 2),
                        date(2020, 1, 3),
                        date(2020, 1, 4),
                    ],
                },
                SparkDataFrameDomain(
                    {
                        "timestamp": SparkTimestampColumnDescriptor(allow_null=False),
                        "date": SparkDateColumnDescriptor(allow_null=False),
                    }
                ),
                {
                    "timestamp": [
                        datetime(2020, 1, 1),
                        datetime(2020, 1, 2),
                        datetime(2020, 1, 3),
                        datetime(2020, 1, 4),
                    ],
                    "date": [
                        date(2020, 1, 1),
                        date(2020, 1, 2),
                        date(2020, 1, 3),
                        date(2020, 1, 4),
                    ],
                },
                SparkDataFrameDomain(
                    {
                        "timestamp": SparkTimestampColumnDescriptor(allow_null=True),
                        "date": SparkDateColumnDescriptor(allow_null=False),
                    }
                ),
                ["timestamp", "date"],
                "outer",
                True,
                {
                    "timestamp": [
                        datetime(2020, 1, 1),
                        datetime(2020, 1, 2),
                        datetime(2020, 1, 3),
                        datetime(2020, 1, 4),
                    ],
                    "date": [
                        date(2020, 1, 1),
                        date(2020, 1, 2),
                        date(2020, 1, 3),
                        date(2020, 1, 4),
                    ],
                },
                SparkDataFrameDomain(
                    {
                        "timestamp": SparkTimestampColumnDescriptor(allow_null=True),
                        "date": SparkDateColumnDescriptor(allow_null=False),
                    }
                ),
            ),
            (  # left join, only using some common columns, weird names
                {
                    "A": [1, 2, 3, 4],
                    "A_left": ["a", "b", "c", "d"],
                    "A_right": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "A_left": SparkStringColumnDescriptor(allow_null=False),
                        "A_right": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"A": [3, 4, 5, 6], "A_left": ["b", "c", "d", "e"]},
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "A_left": SparkStringColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "left",
                False,
                {
                    "A": [1, 2, 3, 4],
                    "A_left_left": ["a", "b", "c", "d"],
                    "A_right": [1.0, 2.0, 3.0, 4.0],
                    "A_left_right": [None, None, "b", "c"],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "A_left_left": SparkStringColumnDescriptor(allow_null=False),
                        "A_right": SparkFloatColumnDescriptor(allow_null=False),
                        "A_left_right": SparkStringColumnDescriptor(allow_null=True),
                    }
                ),
            ),
            (  # outer join, only using some common columns
                {
                    "A": [1, 2, 3, 4],
                    "B": ["a", "b", "c", "d"],
                    "C": [1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                    }
                ),
                {"D": [3, 4, 5, 6], "C": [1.0, 2.0, 3.0, 4.0], "A": [4, 5, 6, 7]},
                SparkDataFrameDomain(
                    {
                        "D": SparkIntegerColumnDescriptor(allow_null=True),
                        "C": SparkFloatColumnDescriptor(allow_null=False),
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                    }
                ),
                ["A"],
                "outer",
                False,
                {
                    "A": [1, 2, 3, 4, 5, 6, 7],
                    "B": ["a", "b", "c", "d", None, None, None],
                    "C_left": [1.0, 2.0, 3.0, 4.0, None, None, None],
                    "D": [None, None, None, 3, 4, 5, 6],
                    "C_right": [None, None, None, 1.0, 2.0, 3.0, 4.0],
                },
                SparkDataFrameDomain(
                    {
                        "A": SparkIntegerColumnDescriptor(allow_null=False),
                        "B": SparkStringColumnDescriptor(allow_null=True),
                        "C_left": SparkFloatColumnDescriptor(allow_null=True),
                        "D": SparkIntegerColumnDescriptor(allow_null=True),
                        "C_right": SparkFloatColumnDescriptor(allow_null=True),
                    }
                ),
            ),
        ]
    )
    def test_correctness(
        self,
        left: Dict[str, List[Any]],
        left_domain: SparkDataFrameDomain,
        right: Dict[str, List[Any]],
        right_domain: SparkDataFrameDomain,
        on: List[str],
        how: str,
        nulls_are_equal: bool,
        expected: Dict[str, List[Any]],
        expected_domain: SparkDataFrameDomain,
    ):
        """Test that join returns the expected result."""

        def to_sdf(data: Dict[str, List[Any]]) -> DataFrame:
            return self.spark.createDataFrame(
                list(zip(*data.values())), schema=list(data)
            )

        left_sdf = to_sdf(left)
        # this is making sure the test input is okay, not the function
        left_domain.validate(left_sdf)
        right_sdf = to_sdf(right)
        right_domain.validate(right_sdf)
        double_check_expected_domain = domain_after_join(
            left_domain, right_domain, on=on, how=how, nulls_are_equal=nulls_are_equal
        )
        assert expected_domain == double_check_expected_domain, str(
            expected_domain
        ) + str(double_check_expected_domain)
        expected_sdf = to_sdf(expected)
        expected_domain.validate(expected_sdf)
        actual = join(
            left=left_sdf,
            right=right_sdf,
            on=on,
            how=how,
            nulls_are_equal=nulls_are_equal,
        )
        self.assert_frame_equal_with_sort(actual.toPandas(), expected_sdf.toPandas())

    def test_left_and_right_are_from_the_same_source(self):
        """Previous implementation got confused when joining a DataFrame with a view."""
        df = self.spark.createDataFrame(
            [("0", 1), ("1", 0), ("1", 2)], schema=["A", "B"]
        )
        left = df
        right = df.filter("B = 2")
        actual = join(left=left, right=right, how="left", nulls_are_equal=True)
        expected = left
        self.assert_frame_equal_with_sort(actual.toPandas(), expected.toPandas())

    @parameterized.expand(COLUMNS_AFTER_JOIN_ERROR_TEST_CASES)
    def test_columns_after_join_validation(
        self,
        left_columns: List[str],
        right_columns: List[str],
        on: Optional[List[str]],
        message: str,
    ):
        """Test that join raises an error when the columns parameters are not valid."""
        left = self.spark.createDataFrame(
            [tuple(range(len(left_columns)))], schema=left_columns
        )
        right = self.spark.createDataFrame(
            [tuple(range(len(right_columns)))], schema=right_columns
        )
        with self.assertRaisesRegex(ValueError, re.escape(message)):
            join(left=left, right=right, on=on)

    @parameterized.expand(DOMAIN_AFTER_JOIN_ERROR_TEST_CASES)
    def test_validation(self, args_updates: Dict[str, Any], message: str):
        """Test that invalid inputs raise an error."""
        domain_args: Any = {
            "left_domain": SparkDataFrameDomain(
                {"column": SparkStringColumnDescriptor()}
            ),
            "right_domain": SparkDataFrameDomain(
                {"column": SparkStringColumnDescriptor()}
            ),
            "on": ["column"],
            "how": "inner",
            "nulls_are_equal": True,
        }
        domain_args.update(args_updates)
        try:
            left = self.spark.createDataFrame(
                [], schema=domain_args["left_domain"].spark_schema
            )
            right = self.spark.createDataFrame(
                [], schema=domain_args["right_domain"].spark_schema
            )
        except AttributeError as e:
            assert re.match("'.*' object has no attribute 'spark_schema'", str(e))
            return

        with self.assertRaisesRegex(Exception, re.escape(message)):
            join(
                left=left,
                right=right,
                on=domain_args["on"],
                how=domain_args["how"],
                nulls_are_equal=domain_args["nulls_are_equal"],
            )



"""Functions for truncating Spark DataFrames."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from typing import List

from pyspark.sql import DataFrame, Window
from pyspark.sql import functions as sf

from tmlt.core.utils.misc import get_nonconflicting_string


def truncate_large_groups(
    df: DataFrame, grouping_columns: List[str], threshold: int
) -> DataFrame:
    """Order rows by a hash function and keep at most `threshold` rows for each group.

    Example:
        ..
            >>> import pandas as pd
            >>> from pyspark.sql import SparkSession
            >>> from tmlt.core.domains.spark_domains import SparkStringColumnDescriptor
            >>> from tmlt.core.utils.misc import print_sdf
            >>> spark = SparkSession.builder.getOrCreate()
            >>> spark_dataframe = spark.createDataFrame(
            ...     pd.DataFrame(
            ...         {
            ...             "A": ["a1", "a2", "a3", "a3", "a3"],
            ...             "B": ["b1", "b1", "b2", "b2", "b3"],
            ...         }
            ...     )
            ... )

        >>> # Example input
        >>> print_sdf(spark_dataframe)
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        >>> print_sdf(truncate_large_groups(spark_dataframe, ["A"], 3))
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        >>> print_sdf(truncate_large_groups(spark_dataframe, ["A"], 2))
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        >>> print_sdf(truncate_large_groups(spark_dataframe, ["A"], 1))
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2

    Args:
        df: DataFrame to truncate.
        grouping_columns: Columns defining the groups.
        threshold: Maximum number of rows to include for each group.
    """
    rank_column = get_nonconflicting_string(df.columns)
    row_index_column = get_nonconflicting_string(df.columns + [rank_column])
    hash_column = get_nonconflicting_string(
        df.columns + [row_index_column, rank_column]
    )
    distinct_row_partitions = Window.partitionBy(*df.columns).orderBy(*df.columns)
    shuffled_partitions = Window.partitionBy(*grouping_columns).orderBy(
        hash_column, *df.columns
    )
    return (
        df.withColumn(row_index_column, sf.row_number().over(distinct_row_partitions))
        .withColumn(
            hash_column,
            sf.hash(*df.columns, row_index_column),  # pylint: disable=no-member
        )
        .withColumn(
            rank_column,
            sf.row_number().over(shuffled_partitions),  # pylint: disable=no-member
        )
        .filter(f"{rank_column}<={threshold}")
        .drop(rank_column, hash_column, row_index_column)
    )


def drop_large_groups(
    df: DataFrame, grouping_columns: List[str], threshold: int
) -> DataFrame:
    """Drop all rows for groups that have more than `threshold` rows.

    Example:
        ..
            >>> import pandas as pd
            >>> from pyspark.sql import SparkSession
            >>> from tmlt.core.domains.spark_domains import SparkStringColumnDescriptor
            >>> from tmlt.core.utils.misc import print_sdf
            >>> spark = SparkSession.builder.getOrCreate()
            >>> spark_dataframe = spark.createDataFrame(
            ...     pd.DataFrame(
            ...         {
            ...             "A": ["a1", "a2", "a3", "a3", "a3"],
            ...             "B": ["b1", "b1", "b2", "b2", "b3"],
            ...         }
            ...     )
            ... )

        >>> # Example input
        >>> print_sdf(spark_dataframe)
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        >>> print_sdf(drop_large_groups(spark_dataframe, ["A"], 3))
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        >>> print_sdf(drop_large_groups(spark_dataframe, ["A"], 2))
            A   B
        0  a1  b1
        1  a2  b1
        >>> print_sdf(drop_large_groups(spark_dataframe, ["A"], 1))
            A   B
        0  a1  b1
        1  a2  b1

    Args:
        df: DataFrame to truncate.
        grouping_columns: Columns defining the groups.
        threshold: Threshold for dropping groups. If more than `threshold` rows belong
            to the same group, all rows in that group are dropped.
    """
    count_column = get_nonconflicting_string(df.columns)
    partitions = Window.partitionBy(*grouping_columns)
    return (
        df.withColumn(
            count_column,
            sf.count(sf.lit(1)).over(partitions),  # pylint: disable=no-member
        )
        .filter(f"{count_column}<={threshold}")
        .drop(count_column)
    )


def limit_keys_per_group(
    df: DataFrame, grouping_columns: List[str], key_columns: List[str], threshold: int
) -> DataFrame:
    """Order keys by a hash function and keep at most `threshold` keys for each group.

    .. note::

        After truncation there may still be an unbounded number of rows per key, but
        at most `threshold` keys per group

    Example:
        ..
            >>> import pandas as pd
            >>> from pyspark.sql import SparkSession
            >>> from tmlt.core.domains.spark_domains import SparkStringColumnDescriptor
            >>> from tmlt.core.utils.misc import print_sdf
            >>> spark = SparkSession.builder.getOrCreate()
            >>> spark_dataframe = spark.createDataFrame(
            ...     pd.DataFrame(
            ...         {
            ...             "A": ["a1", "a2", "a3", "a3", "a3", "a4", "a4", "a4"],
            ...             "B": ["b1", "b1", "b2", "b2", "b3", "b1", "b2", "b3"],
            ...         }
            ...     )
            ... )

        >>> # Example input
        >>> print_sdf(spark_dataframe)
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        5  a4  b1
        6  a4  b2
        7  a4  b3
        >>> print_sdf(
        ...     limit_keys_per_group(
        ...         df=spark_dataframe,
        ...         grouping_columns=["A"],
        ...         key_columns=["B"],
        ...         threshold=2,
        ...     )
        ... )
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a3  b3
        5  a4  b2
        6  a4  b3
        >>> print_sdf(
        ...     limit_keys_per_group(
        ...         df=spark_dataframe,
        ...         grouping_columns=["A"],
        ...         key_columns=["B"],
        ...         threshold=1,
        ...     )
        ... )
            A   B
        0  a1  b1
        1  a2  b1
        2  a3  b2
        3  a3  b2
        4  a4  b3

    Args:
        df: DataFrame to truncate.
        grouping_columns: Columns defining the groups.
        key_columns: Column defining the keys.
        threshold: Maximum number of keys to include for each group.
    """
    rank_column = get_nonconflicting_string(df.columns)
    hash_column = get_nonconflicting_string(df.columns + [rank_column])
    shuffled_partitions = Window.partitionBy(*grouping_columns).orderBy(
        hash_column, *key_columns
    )
    return (
        df.withColumn(
            hash_column, sf.hash(*grouping_columns, *key_columns)
        )  # pylint: disable=no-member
        .withColumn(
            rank_column,
            sf.dense_rank().over(shuffled_partitions),  # pylint: disable=no-member
        )
        .filter(f"{rank_column}<={threshold}")
        .drop(rank_column, hash_column)
    )



"""Benchmarking script for spark-based count and sum aggregations."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2023

from math import log
from random import randint
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from pyspark.sql import SparkSession
from pyspark.sql import functions as sf
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import LongType, StructField, StructType

from benchmarking_utils import Timer, write_as_html
from tmlt.core.domains.spark_domains import (
    SparkDataFrameDomain,
    SparkIntegerColumnDescriptor,
)
from tmlt.core.metrics import SymmetricDifference
from tmlt.core.transformations.spark_transformations.agg import (
    create_count_aggregation,
    create_sum_aggregation,
)
from tmlt.core.transformations.spark_transformations.groupby import (
    create_groupby_from_column_domains,
)
from tmlt.core.utils.exact_number import ExactNumberInput
from tmlt.core.utils.testing import PySparkTest


def evaluate_runtime(
    groupby_domains: Dict[str, List[int]],
    dataframe: DataFrame,
    input_domain: SparkDataFrameDomain,
    measure_column: str,
    low: ExactNumberInput,
    high: ExactNumberInput,
) -> Tuple[float, float, float]:
    """Returns the runtimes for count, sum, and both aggregations."""
    groupby = create_groupby_from_column_domains(
        input_domain=input_domain,
        input_metric=SymmetricDifference(),
        use_l2=False,
        column_domains=groupby_domains,
    )
    count = create_count_aggregation(
        input_domain=groupby.output_domain,
        input_metric=groupby.output_metric,
    )
    groupby_count = groupby | count
    with Timer() as count_timer:
        groupby_count(dataframe).toPandas()
    count_time = count_timer.elapsed

    sum = create_sum_aggregation(
        input_domain=groupby.output_domain,
        input_metric=groupby.output_metric,
        measure_column=measure_column,
        lower=low,
        upper=high,
        sum_column=f"sum{measure_column}",
    )
    groupby_sum = groupby | sum
    with Timer() as sum_timer:
        groupby_sum(dataframe).toPandas()
    sum_time = sum_timer.elapsed
    return round(count_time, 3), round(sum_time, 3)


def main():
    """Evaluate count and sum runtimes for different group counts and sizes."""
    spark = SparkSession.builder.getOrCreate()
    benchmark_result = pd.DataFrame(
        [],
        columns=[
            "domain_size",
            "group_size",
            "group_count",
            "num_records",
            "num_groupby_columns",
            "count_time (s)",
            "sum_time (s)",
        ],
    )
    input_domain = SparkDataFrameDomain(
        {"A": SparkIntegerColumnDescriptor(), "X": SparkIntegerColumnDescriptor()}
    )

    # Runtimes on Empty DataFrame
    empty_df = spark.createDataFrame([], schema=input_domain.spark_schema)
    for domain_size in [100, 400, 10000, 40000, 160000, 640000]:
        count_time, sum_time = evaluate_runtime(
            groupby_domains={"A": list(range(domain_size))},
            dataframe=empty_df,
            input_domain=input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": domain_size,
            "group_size": 0,
            "group_count": domain_size,
            "num_records": 0,
            "num_groupby_columns": 1,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # Single Groupby Column of varying domain sizes (1 row/group)
    for domain_size in [100, 400, 10000, 40000, 160000, 640000]:
        df = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                [(i, randint(0, 1)) for i in range(domain_size)]
            ),
            schema=input_domain.spark_schema,
        )
        count_time, sum_time = evaluate_runtime(
            groupby_domains={"A": list(range(domain_size))},
            dataframe=df,
            input_domain=input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": domain_size,
            "group_size": 1,
            "group_count": domain_size,
            "num_records": domain_size,
            "num_groupby_columns": 1,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # Single groupby column, group size = 1M
    for size in [100000, 900000, 10000000]:
        df = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                [
                    (i, randint(0, 1))
                    for _ in range(100000)
                    for i in range(int(size / 100000))
                ]
            ),
            schema=input_domain.spark_schema,
        )
        count_time, sum_time = evaluate_runtime(
            {"A": list(range(int(size / 100000)))},
            df,
            input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": int(size / 100000),
            "group_size": 100000,
            "group_count": int(size / 100000),
            "num_records": size,
            "num_groupby_columns": 1,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # Group size = 10K
    for size in [10000, 100000, 1000000, 10000000]:
        df = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                [
                    (i, randint(0, 1))
                    for j in range(10000)
                    for i in range(int(size / 10000))
                ]
            ),
            schema=input_domain.spark_schema,
        )
        count_time, sum_time = evaluate_runtime(
            groupby_domains={"A": list(range(int(size / 10000)))},
            dataframe=df,
            input_domain=input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": int(size / 10000),
            "group_size": 10000,
            "group_count": int(size / 10000),
            "num_records": size,
            "num_groupby_columns": 1,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # Group size = 100
    for size in [10000, 40000, 160000, 640000, 2560000]:
        df = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                [(i, randint(0, 1)) for j in range(100) for i in range(int(size / 100))]
            ),
            schema=input_domain.spark_schema,
        )
        count_time, sum_time = evaluate_runtime(
            groupby_domains={"A": list(range(int(size / 100)))},
            dataframe=df,
            input_domain=input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": int(size / 100),
            "group_size": 100,
            "group_count": int(size / 100),
            "num_records": size,
            "num_groupby_columns": 1,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # Multiple groupby columns
    domain_size = 2
    group_size = 1
    for num_cols in [1, 2, 4, 8, 12, 16, 18, 20]:
        group_count = domain_size ** num_cols
        num_records = group_count * group_size
        groupby_columns = ["Col_{}".format(i) for i in range(num_cols)]
        columns = groupby_columns + ["X"]
        input_domain = SparkDataFrameDomain(
            dict.fromkeys(columns, SparkIntegerColumnDescriptor())
        )
        schema = StructType(
            [
                StructField("Col_{}".format(i), LongType(), True)
                for i in range(num_cols)
            ]
        )
        sdf = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                np.repeat(
                    np.transpose(
                        np.unravel_index(
                            range(group_count), shape=(domain_size,) * num_cols
                        )
                    ),
                    group_size,
                    axis=0,
                ).tolist()
            ),
            schema=schema,
        )
        sdf = sdf.withColumn("X", sf.lit(100))
        groupby_domains = dict.fromkeys(groupby_columns, list(range(domain_size)))
        count_time, sum_time = evaluate_runtime(
            groupby_domains,
            sdf,
            input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": 2,
            "group_size": group_size,
            "group_count": group_count,
            "num_records": num_records,
            "num_groupby_columns": num_cols,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    # various domain sizes and columns
    group_size = 100
    group_count = 2 ** 8
    num_records = group_count * group_size
    for domain_size in [256, 16, 4, 2]:
        num_cols = int(log(group_count, domain_size))
        groupby_columns = ["Col_{}".format(i) for i in range(num_cols)]
        columns = groupby_columns + ["X"]
        input_domain = SparkDataFrameDomain(
            dict.fromkeys(columns, SparkIntegerColumnDescriptor())
        )
        schema = StructType(
            [
                StructField("Col_{}".format(i), LongType(), True)
                for i in range(num_cols)
            ]
        )
        sdf = spark.createDataFrame(  # pylint: disable=no-member
            spark.sparkContext.parallelize(
                np.repeat(
                    np.transpose(
                        np.unravel_index(
                            range(group_count), shape=(domain_size,) * num_cols
                        )
                    ),
                    group_size,
                    axis=0,
                ).tolist()
            ),
            schema=schema,
        )
        sdf = sdf.withColumn("X", sf.lit(100))
        groupby_domains = dict.fromkeys(groupby_columns, list(range(domain_size)))
        count_time, sum_time = evaluate_runtime(
            groupby_domains,
            sdf,
            input_domain,
            measure_column="X",
            low=0,
            high=1,
        )
        row = {
            "domain_size": domain_size,
            "group_size": group_size,
            "group_count": group_count,
            "num_records": num_records,
            "num_groupby_columns": num_cols,
            "count_time (s)": count_time,
            "sum_time (s)": sum_time,
        }
        benchmark_result = benchmark_result.append(row, ignore_index=True)

    write_as_html(benchmark_result, "count_sum.html")


if __name__ == "__main__":
    PySparkTest.setUpClass()
    main()
    PySparkTest.tearDownClass()



"""Abstract class for testing metrics."""

# SPDX-License-Identifier: Apache-2.0
# Copyright Tumult Labs 2022

import copy

# pylint: disable=unused-import
from abc import ABC, abstractmethod
from contextlib import nullcontext as does_not_raise
from typing import Any, Callable, ContextManager, Dict, Optional, Type

import pytest

from tmlt.core.domains.base import Domain
from tmlt.core.metrics import Metric
from tmlt.core.utils.exact_number import ExactNumberInput
from tmlt.core.utils.testing import assert_property_immutability, get_all_props


class MetricTests(ABC):
    """Test classes for Metrics should inherit from this class."""

    @abstractmethod
    @pytest.fixture
    def metric_type(self) -> Type[Metric]:  # pylint: disable=no-self-use
        """Returns the type of the metric to be tested."""
        return Metric

    @abstractmethod
    def test_mutable_inputs(
        self,
        metric_type: Type[Metric],
        metric_args: Dict[str, Any],
        key: str,
        mutator: Callable,
    ):
        """The mutable inputs to the metric are copied.

        Args:
            metric_type: The type of metric to be constructed.
            metric_args: The arguments to the metric.
            key: The parameter name to be changed.
            mutator: A lambda function that mutates the parameter.
        """
        metric = metric_type(**metric_args)
        assert hasattr(metric, key)
        expected = copy.deepcopy(getattr(metric, key))
        mutator(metric_args[key])
        actual_value = getattr(metric, key)
        assert (
            getattr(metric, key) == expected
        ), f"Expected {key} to be {expected}, got {actual_value}"

    @abstractmethod
    def test_property_immutability(self, metric: Metric):
        """The properties return copies for mutable values.

        Args:
            metric: The measure to be tested.
        """
        props = [prop[0] for prop in get_all_props(type(metric))]
        for prop in props:
            assert_property_immutability(metric, prop)

    @abstractmethod
    def test_properties(self, metric_type: type, metric_args: Dict[str, Any]):
        """All properties have the expected values.

        Args:
            metric_type: The type of metric to be constructed.
            metric_args: The arguments to the metric.
        """
        metric = metric_type(**metric_args)
        for arg_name, value in metric_args.items():
            if not hasattr(metric, arg_name):
                continue
            assert getattr(metric, arg_name) == value

    @abstractmethod
    def test_construct_component(
        self,
        metric_type: Type[Metric],
        metric_args: Dict[str, Any],
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Initialization behaves correctly.

        The metric is constructed correctly and raises exceptions
        when initialized with invalid inputs.

        Args:
            metric_type: The type of metric to be constructed.
            metric_args: The arguments to the metric.
            expectation: A context manager that captures the
                correct expected type of error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            metric_type(**metric_args)
        if exception_properties is None or len(exception_properties) == 0:
            return
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception, prop), f"{prop} not in {exception}"
            actual_value = getattr(exception, prop)
            assert (
                getattr(exception, prop) == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_eq(self, metric: Metric, other_metric: Metric, expected: bool):
        """Metrics can be compared correctly.

        Args:
            metric: The metric to compare.
            other_metric: The metric to compare with.
            expected: Expected equivalence.
        """
        assert (metric == other_metric) == expected

    @abstractmethod
    def test_validate(
        self,
        metric: Metric,
        candidate: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Validate works correctly.

        Args:
            metric: The metric to test.
            candidate: The value to validate using metric.
            expectation: A context manager that captures the
                correct expected type of error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            metric.validate(candidate)
        if exception_properties is None or len(exception_properties) == 0:
            return
        # Help out mypy
        assert isinstance(exception, pytest.ExceptionInfo)
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception.value, prop), f"Expected prop was missing: {prop}"
            actual_value = getattr(exception.value, prop)
            assert (
                actual_value == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_compare(
        self,
        metric: Metric,
        value1: ExactNumberInput,
        value2: ExactNumberInput,
        expected: bool,
    ):
        """Compare returns the expected result.

        Args:
            metric: The metric doing the comparison.
            value1: A distance between two datasets under this metric.
            value2: A distance between two datasets under this metric.
            expected: Expected equivalence.
        """
        assert metric.compare(value1, value2) == expected

    @abstractmethod
    def test_repr(self, metric: Metric, representation: str):
        """`repr` creates the correct representation of the metric.

        Args:
            metric: The metric to be represented.
            representation: The expected string representation of the
            metric.
        """
        assert repr(metric) == representation

    @abstractmethod
    def test_distance(
        self,
        metric: Metric,
        domain: Domain,
        value1: Any,
        value2: Any,
        expectation: ContextManager[None],
        exception_properties: Optional[Dict[str, Any]],
    ):
        """Distances raises an exception for invalid inputs.

        Args:
            metric: The metric performing the comparison.
            domain: The domain under which the metric performs
                the comparison under.
            value1: A distance between two datasets under this metric.
            value2: A distance between two datasets under this metric.
            expectation: A context manager that captures the
                correct expected type of error that is raised.
            exception_properties: A dictionary containing all the property:value pairs
                the exception is expected to have. Mostly used for testing the custom
                exceptions.
        """
        with expectation as exception:
            metric.distance(value1, value2, domain)
        if exception_properties is None or len(exception_properties) == 0:
            return
        for prop, expected_value in exception_properties.items():
            assert hasattr(exception, prop), f"{prop} not in {exception}"
            actual_value = getattr(exception, prop)
            assert (
                getattr(exception, prop) == expected_value
            ), f"Expected {prop} to be {expected_value}, got {actual_value}"

    @abstractmethod
    def test_supports_domain(self, metric: Metric, domain: Domain, expected: bool):
        """Metric supports the correct domains.

        Args:
            metric: The metric to test.
            domain: The domain to test.
            expected: The expected result.
        """
        assert metric.supports_domain(domain) == expected
